--[[
Name: LibGroupTalents-1.0
Revision: $Rev: 13 $
Author: Zek
Documentation: http://wowace.com/wiki/LibGroupTalents-1.0
SVN: svn://svn.wowace.com/wow/libgrouptalents-1-0/mainline/trunk
Description: Talent Library abstraction layer to provide easy interface to the lower level system
Dependancies: LibStub, CallbackHandler-1.0, LibTalentQuery-1.0
License: GPL v3

Purpose:
	LibGroupTalents-1.0 is intended to do the following basic functions usually handled at the mod level.

	- Maintain a raid wide table of talents, automatically updated on roster changes, notifying you on talent receipts.
	- Provide easy access to talent queries (spec weight, spec name, specific talent presence)
	- Monitor talent changes in players, and notify of changes (respec, talent swap, update after out of sight, level up)
	- Monitor player roles, and notify of changes (melee, tank, healer, caster)
	- Communicate directly with itself to other users to update talents via addon channel when possible

Notes:
	The LibTalentQuery-1.0 dependancy must be included before LibGroupTalents-1.0 in any lib.xml or mod side TOC declarations.

Functions:
	UnitHasTalent(unit, talentName[, group])-- Returns: Points spent in talent or nil
	GUIDHasTalent(guid, talentName[, group])-- As UnitHasTalent
	GetUnitTalentSpec(unitid[, group])		-- Returns: Dominant Tree, spent1, spent2, spent3
	GetGUIDTalentSpec(guid[, group])		-- As GetUnitTalentSpec
	GetUnitTalents(unit, refresh)			-- Returns: Raw talent information in form of table of 3 strings of points spent. The refresh arg will force a re-query of the unit's talents
	GetGUIDTalents(guid, refresh)			-- As GetUnitTalents
	GetUnitRole(unit)						-- Returns one of: "melee", "caster", "healer", "tank"
	GetGUIDRole(guid)						-- As GetUnitRole
	RefreshTalentsByUnit(unit)				-- Force a refresh of talents for the specific unit
	RefreshTalentsByGUID(guid)				-- Force a refresh of talents for the specific player GUID
	GetTreeNames(class)						-- Returns: The three talent tree names for that class (Note: These return values are only valid after a player of that class has been inspected)
	GetTreeIcons(class)						-- Returns: The three talent tree icons for that class (Note: As above)
	GetTalentCount()						-- Returns: Talent info got, Talent info missing
	GetTalentMissingNames()					-- Returns: Comma delimited list of player names we're missing talents for
	GetClassTalentInfo(class, talentName)	-- Returns: Max Rank, Icon, Tab, Tier, Column, Tree Index
	GetUnitStorageString(unit)				-- Returns: An encoded data string containing talent information for the player which can be stored by mods to set in later sessions using SetStorageString()
	GetGUIDStorageString(guid)				-- As GetUnitStorageString
	SetStorageString(talentString)			-- Returns: true on success (applicable). Any second return value indicates the data was invalid and should not be kept
	GetUnitGlyphs(unit, group)				-- Returns: Up to 6 spell IDs for the currently assigned Glyphs (Note: For the moment, we can only see the glyphs of players running LibGroupTalents-1.0)
	GetGUIDGlyphs(guid, group)				-- As GetUnitGlyphs
	UnitHasGlyph(unit, glyph [, group])		-- Returns: true if the player has the glyph associated with spellID or spellName (Note: For the moment, we can only see the glyphs of players running LibGroupTalents-1.0)
	GUIDHasGlyph(unit, glyph [, group])		-- As UnitHasGlyph
	PurgeAndRescanTalents()					-- Wipe current roster of all talents and rescan from start

Convenience Functions (Similar to Blizzard API functions, but callable with a unit ID):
	GetActiveTalentGroup(unit)				-- Returns: Active talent group for unit
	GetNumTalentGroups(unit)				-- Returns: Number of talent groups for unit
	GetNumTalentTabs(unit)					-- Returns: Number of talent tabs. Here's a clue; it's going to be 3...
	GetTalentTabInfo(unit, tab[, group])	-- Returns: Tree Name, Tree Icon, Points Spent, Tree Background
	GetNumTalents(unit, tab)				-- Returns: Number of talents for specified tree
	GetTalentInfo(unit, tab, index[, group])-- Returns: Talent Name, Icon, Tier, Column, Points Spent, Max Rank (Note that preview return values are not given unless called with "player")
	GetUnspentTalentPoints(unit[, group])	-- Returns: Number of un-spent talent points for the unit

Events:
	LibGroupTalents_Update(guid, unit, newSpec, n1, n2, n3 [, oldSpec, o1, o2, o3])	-- Received updated talents. If it's a respec, or old set is know, it passes the old info also (this is not sent if new talent scan is same as previous)
	LibGroupTalents_UpdateComplete(guid1, guid2[, ...])	-- Sent when there are no more pending talent reads due (passes all GUIDs that were updated since last time this event was fired)
	LibGroupTalents_Add(guid, unit, name, realm)	-- Unit added to talent roster (Talents not necessarily available yet, but this is the mod's chance to feed talents using SetStorageString)
	LibGroupTalents_Remove(guid, name, realm)										-- Unit removed from talent roster (This is your last chance to store talents if required using GetUnitStorageString)
	LibGroupTalents_RoleChange(guid, unit, newrole, oldrole)		-- Roles are: "melee", "caster", "healer", "tank"
	LibGroupTalents_GlyphUpdate(guid, unit)			-- Fired when a player's glyphs change (Note: For the moment, we can only see the glyphs of players running LibGroupTalents-1.0)

]]

local MAJOR, MINOR = "LibGroupTalents-1.0", tonumber(("$Rev: 13 $"):match("(%d+)"))
local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

--@debug@
LGTD = true
local function d(...)
	if (LGTD) then
		ChatFrame1:AddMessage("|cFF8080FFLGT: |r"..format(...))
	end
end
LGT = lib
--@no-debug@

local TalentQuery = LibStub("LibTalentQuery-1.0")
local ChatThrottleLib = _G.ChatThrottleLib

lib.roster = lib.roster or {}
lib.classTalentData = lib.classTalentData or {}
lib.batch = lib.batch or {}

local specChangers = {
	[GetSpellInfo(63645)] = 1,					-- Activate Primary Spec
	[GetSpellInfo(63644)] = 2,					-- Activate Secondary Spec
}

local frame = lib.frame
if (not frame) then
	frame = CreateFrame("Frame", "LibGroupTalents_Frame")
	lib.frame = frame
end
frame:UnregisterAllEvents()
frame:RegisterEvent("RAID_ROSTER_UPDATE")
frame:RegisterEvent("PARTY_MEMBERS_CHANGED")
frame:RegisterEvent("UNIT_NAME_UPDATE")
frame:RegisterEvent("PLAYER_TALENT_UPDATE")
frame:RegisterEvent("UNIT_LEVEL")
frame:RegisterEvent("UNIT_AURA")					-- Always get a UNIT_AURA when a unit's UnitIsVisible() changes
frame:RegisterEvent("CHAT_MSG_ADDON")
frame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
frame:RegisterEvent("PLAYER_LOGIN")
frame:RegisterEvent("GLYPH_ADDED")
frame:RegisterEvent("GLYPH_REMOVED")
frame:RegisterEvent("GLYPH_UPDATED")

frame:SetScript("OnEvent", function(self, event, ...)
	return lib[event](lib, ...)
end)

if not lib.events then
	lib.events = LibStub("CallbackHandler-1.0"):New(lib)
end

local next, select, pairs, type = next, select, pairs, type
local new, del, deepDel
do
	local list = setmetatable({},{__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if (t) then
			wipe(t)
			t[''] = true
			t[''] = nil
			list[t] = true
		end
	end
	function deepDel(t)
		if (t) then
			for k,v in pairs(t) do
				if type(v) == "table" then
					deepDel(v)
				end
				t[k] = nil
			end
			t[''] = true
			t[''] = nil
			list[t] = true
		end
	end
end

do
	local delay = 0
	frame:SetScript("OnUpdate", function(self, elapsed)
		if (lib.raidRosterUpdate) then
			lib.raidRosterUpdate = nil
			lib:OnRaidRosterUpdate()
		end

		if (lib.refreshCheckTimer) then
			lib.refreshCheckTimer = lib.refreshCheckTimer - elapsed
			if (lib.refreshCheckTimer < 0) then
				lib.refreshCheckTimer = nil
				lib:CheckForMissingTalents()
			end
		end

		if (lib.talentTimers) then
			delay = delay + elapsed
			if (delay > 1) then
				delay = 0
				local now = GetTime()
				local triggers
				for guid,when in pairs(lib.talentTimers) do
					if (now > when) then
						-- Pass to second table to process, because RefreshTimers can affect this talentTimers table
						-- So it's important we're not still iterating it at the time
						if (not triggers) then
							triggers = new()
						end
						triggers[guid] = true
						lib.talentTimers[guid] = nil
						if (not next(lib.talentTimers)) then
							lib.talentTimers = del(lib.talentTimers)
							break
						end
					end
				end

				if (triggers) then
					for guid in pairs(triggers) do
						lib:RefreshTalentsByGUID(guid)
					end
					del(triggers)
				end
			end
		end

		if (not lib.talentTimers and not lib.refreshCheckTimer) then
			self:Hide()
		end
	end)
end
frame:Show()
lib.raidRosterUpdate = true

-- GetGUIDTalentsRaw
local function GetGUIDTalentsRaw(guid, group)
	local r = guid and lib.roster[guid]
	if (r) then
		local talents = r.talents
		if (group and group ~= r.active) then
			return r.alternateTalents[group], r
		else
			return r.talents, r
		end
	end
end

-- PLAYER_LOGIN
function lib:PLAYER_LOGIN()
	ChatThrottleLib = _G.ChatThrottleLib
	lib.PLAYER_LOGIN = nil
end

-- RAID_ROSTER_UPDATE
function lib:RAID_ROSTER_UPDATE()
	self.raidRosterUpdate = true
	frame:Show()
end
lib.PARTY_MEMBERS_CHANGED = lib.RAID_ROSTER_UPDATE

-- UNIT_NAME_UPDATE
function lib:UNIT_NAME_UPDATE(unit)
	local guid = unit and UnitGUID(unit)
	local r = guid and self.roster[guid]
	if (r) then
		r.name, r.realm = UnitName(unit)
		if (r.realm == "") then
			r.realm = nil
		end
		r.level = UnitLevel(unit)
		if (not r.talents) then
			self:CheckForMissingTalents()
		end
	end
end

-- AnyPending
local function AnyPending()
	local checkUpdate
	for guid,info in pairs(lib.roster) do
		if (UnitIsConnected(info.name)) then
			if (lib.wasOffline) then
				lib.wasOffline[guid] = nil
			end
			if (not info.talents or info.refresh) then
				return true
			end
		else
			if (not lib.wasOffline) then
				lib.wasOffline = new()
			end
			lib.wasOffline[guid] = true
		end
	end
	if (lib.wasOffline and not next(lib.wasOffline)) then
		lib.wasOffline = del(lib.wasOffline)
	end
end

-- CheckForUpdateComplete
local function CheckForUpdateComplete()
	-- When all pending updates are complete, send an event to notify nothing else is due
	if (next(lib.batch)) then
		if (not AnyPending()) then
			--d("Fire: LibGroupTalents_UpdateComplete(%s)", table.concat(lib.batch, ", "))
			lib.events:Fire("LibGroupTalents_UpdateComplete", unpack(lib.batch))
			wipe(lib.batch)
		end
	end
end

-- UNIT_LEVEL
function lib:UNIT_LEVEL(unit)
	if (UnitInParty(unit) or UnitInRaid(unit)) then
		self:RefreshTalentsByUnit(unit)
	end
end

-- UNIT_AURA
function lib:UNIT_AURA(unit)
	local guid = UnitGUID(unit)
	local r = guid and self.roster[guid]
	if (r) then
		if (not UnitIsVisible(unit) or (self.wasOffline and self.wasOffline[guid])) then
			if (not self.outOfSight) then
				self.outOfSight = {}
			end
			self.outOfSight[UnitGUID(unit)] = true
			self:RefreshTalentsByUnit(unit)
		end

		if (r.class == "DRUID" or r.class == "DEATHKNIGHT") then
			self:GetGUIDRole(guid, true)		-- Check for role change based on auras (bear form etc)
		end
	end
end

-- OnRaidRosterUpdate
function lib:OnRaidRosterUpdate()
	if (select(2, IsInInstance()) == "pvp") then
		self.distribution = "BATTLEGROUND"
	else
		if (GetNumRaidMembers() > 0) then
			self.distribution = "RAID"
		elseif (GetNumPartyMembers() > 0) then
			self.distribution = "PARTY"
		else
			self.distribution = nil
		end
	end
	if (self.distribution) then
		if (not self.sentHello or (self.sentHello < GetTime() - 15 and not self.users)) then
			self.sentHello = GetTime()
			self:SendCommMessage("HELLO "..MINOR, nil, self.distribution)
		end
	else
		self.sentHello = nil
		self.talentThrottle = del(self.talentThrottle)
	end

	-- Now check for roster changes
	local subtractions = new()
	local additions = new()
	local changes = new()

	if (self.roster) then
		for guid,info in pairs(self.roster) do
			subtractions[guid] = info.level or 0
		end
	end

	for unit in self:IterateRoster() do
		local guid = UnitGUID(unit)
		if (guid) then
			local n = self.roster[guid]
			if (not n) then
				n = new()
				self.roster[guid] = n
			end

			n.name, n.realm = UnitName(unit)
			if (n.realm == "") then
				n.realm = nil				-- Fix this already..
			end
			n.level = UnitLevel(unit)
			n.class = select(2, UnitClass(unit))
			n.unit = unit

			if (subtractions[guid]) then
				if (subtractions[guid] ~= n.level) then
					changes[guid] = unit				-- Level changed, needs a rescan
				end

				subtractions[guid] = nil
			else
				--d("Fire: LibGroupTalents_Add(%s, %s, %s, %s)", guid or "nil", unit or "nil", n.name or "nil", n.realm or "nil")
				self.events:Fire("LibGroupTalents_Add", guid, unit, n.name, n.realm)
				additions[guid] = unit
			end
		end
	end

	if (next(additions)) then
		for guid,unit in pairs(additions) do
			self:GetUnitTalents(unit)
		end
	end

	if (next(changes)) then
		for guid,unit in pairs(changes) do
			self:GetUnitTalents(unit)
		end
	end

	if (next(subtractions)) then
		for guid in pairs(subtractions) do
			local r = self.roster[guid]
			if (r) then
				--d("Fire: LibGroupTalents_Remove(%s, %s, %s)", guid or "nil", r.name or "nil", r.realm or "nil")
				self.events:Fire("LibGroupTalents_Remove", guid, r.name, r.realm)
				self.roster[guid] = deepDel(r)
			end
		end

		CheckForUpdateComplete()
	end

	del(additions)
	del(subtractions)

	self:CheckForMissingTalents()
end

-- ValidateUnit
local function ValidateUnit(r, guid)
	local unit = r.unit	
	if (UnitGUID(unit) ~= guid) then
		local name = r.name .. (r.realm and "-" or "") .. (r.realm or "")
		local index = UnitInRaid(name)
		if (index) then
			r.unit = "raid"..index
			return true
		else
			if (UnitGUID("player") == guid) then
				r.unit = "player"
				return true

			elseif (UnitInParty(name)) then
				for i = 1,4 do
					if (UnitGUID("party"..i) == guid) then
						r.unit = "party"..i
						return true
					end
				end
			end
		end
		return
	end

	return true
end

-- CountTree
local function CountTree(branch)
	local count = 0
	for i = 1,#branch do
		count = count + branch:byte(i) - 48
	end
	return count
end

-- TalentWeight
local function TalentWeight(talents, class)
	if (talents and #talents == 3 and class) then
		local c1, c2, c3 = CountTree(talents[1]), CountTree(talents[2]), CountTree(talents[3])

		local weight = 1
		if (c2 > c1 and c2 > c3) then
			weight = 2
		elseif (c3 > c1 and c3 > c2) then
			weight = 3
		end

		local data = lib.classTalentData[class]
		if (data and data[weight]) then
			return data[weight].name, c1, c2, c3
		end

		return weight, c1, c2, c3
	end
end

do
-- First segment: Player ID (from GUID), Name, level, class, activePage, TalentString()
-- Subsequent: spec number, talentString()

	-- TalentString
	local function TalentString(talents)
		return table.concat(talents, "-")
	end

	-- crc
	local function crc32(str)
		local val = tonumber((select(2, GetBuildInfo())))	-- Use WoW build as CRC base
		for i = 1,#str do
			val = bit.band(val * 2 + str:byte(i), 0xFFFF)
		end
		return val
	end

	-- GetUnitStorageString
	function lib:GetUnitStorageString(unit)
		return self:GetGUIDStorageString(UnitGUID(unit))
	end

	-- GetGUIDStorageString
	-- Make a storage string for mods to store talents.
	-- Rules: 1) Your own realm only 2) Their talents are complete (nothing unspent)
	function lib:GetGUIDStorageString(guid)
		local r = self.roster[guid]
		if (r) then
			local id = tonumber(guid:sub(6), 16)
			if (r.talents and r.active and not r.realm and (not r.unspent or not r.unspent[r.active])) then
				local str = format("%X,%s,%s,%s,%s,%s", id, r.name, r.level, r.class, r.active or 1, TalentString(r.talents))
				if (r.alternateTalents) then
					for i,tree in pairs(r.alternateTalents) do
						str = str .. format(";%d,%s", i, TalentString(tree))
					end
				end
				return str .. ";" .. crc32(str)
			end
		end
	end

	-- SetStorageString
	function lib:SetStorageString(str)
		local ret, retInfo
		if (str) then
			local parts = new(strsplit(";", str))
			if (#parts >= 2) then
				local strCRC = tonumber(parts[#parts])
				local temp = table.concat(parts, ";", 1, #parts - 1)
				if (crc32(temp) == strCRC) then
					local part1 = new(strsplit(",", parts[1]))

					while true do
						local guid = format("0x010%013X", tonumber(part1[1], 16))
						local name = part1[2]
						local level = tonumber(part1[3])
						local class = part1[4]

						local r = self.roster[guid]
						if (not r) then
							error("Unexpected SetStorageString for %s", name)
						end
						if (r.talents) then
							-- We've already received talents for this player
							break
						end

						if (r.level ~= level) then
							-- Won't accept talents for mismatched levels
							retInfo = "Wrong level"
							break
						end
						if (r.class ~= class) then
							-- Class doesn't match, probably a char delete/remake or xrealm
							retInfo = "Wrong class"
							break
						end

						local active = tonumber(part1[5])
						local talents = new(strsplit("-", part1[6]))

						r.talents = talents
						r.active = active
						r.refresh = true

						-- Now alternateTalents
						local failAlt
						for i = 2,#parts - 1 do
							local partN = new(strsplit(",", parts[i]))
							if (#partN == 2) then
								local specNumber = tonumber(partN[1])
								local specTalents = new(strsplit("-", part1[2]))

								if (specNumber) then
									if (not r.alternateTalents) then
										r.alternateTalents = new()
									end
									r.alternateTalents[specNumber] = specTalents
								else
									del(specTalents)
									failAlt = true
									retInfo = "Invalid alternate specs"
									break
								end
							end
						end

						if (not failAlt) then
							ValidateUnit(r, guid)
							local newSpec, n1, n2, n3 = TalentWeight(r.talents, r.class)
							--d("Fire: LibGroupTalents_Update(%s, %s, %s, %s, %s, %s)", guid or "nil", unit and UnitName(unit) or "nil", newSpec or "nil", n1 or "nil", n2 or "nil", n3 or "nil")
							self.events:Fire("LibGroupTalents_Update", guid, r.unit, newSpec, n1, n2, n3)
							self:GetGUIDRole(guid, true)
							ret = true
						end
						break
					end

					del(part1)
				else
					retInfo = "Invalid string"
				end
			end
			del(parts)
		end

		return ret, retInfo
	end
end

-- GetClassTalentData
-- Builds an internal table for talent name -> tree/index lookups.
function lib:GetClassTalentData(unit)
	local _, class = UnitClass(unit)
	if (class) then
		local data = self.classTalentData[class]
		if (not data) then
			local isnotplayer = not UnitIsUnit("player", unit)
			if (GetNumTalentTabs(isnotplayer) > 0) then
				data = new()

				for tab = 1, GetNumTalentTabs(isnotplayer) do
					local tree = new()
					local _
					tree.name, tree.icon, _, tree.background = GetTalentTabInfo(tab, isnotplayer)
					tinsert(data, tree)

					tree.list = new()
					for i = 1,GetNumTalents(tab, isnotplayer) do
						local name, icon, tier, column, currentRank, maxRank = GetTalentInfo(tab, i, isnotplayer)
						if (name) then
							local entry = new()
							entry.name = name
							entry.icon = icon
							entry.tier = tier
							entry.column = column
							entry.maxRank = maxRank
							entry.index = i
							entry.treeIndex = tab
							tinsert(tree.list, entry)
							if (not data.list) then
								data.list = new()
							end
							data.list[name] = entry
						end
					end
				end

				if (next(data)) then
					self.classTalentData[class] = data
				else
					deepDel(data)
				end
			end
		end
	end
end

-- GetTreeNames
function lib:GetTreeNames(class)
	local info = self.classTalentData[class]
	if (info) then
		return info[1].name, info[2].name, info[3].name
	end
end

-- GetTreeIcons
function lib:GetTreeIcons(class)
	local info = self.classTalentData[class]
	if (info) then
		return info[1].icon, info[2].icon, info[3].icon
	end
end

-- ReadTalentGroup
local function ReadTalentGroup(isnotplayer, group, class)
	local numTabs = GetNumTalentTabs(isnotplayer)
	if (numTabs and numTabs >= 3 and GetNumTalents(1, isnotplayer) > 0) then
		local ctd = lib.classTalentData[class]
--@debug@
assert(ctd and ctd[1] and ctd[2] and ctd[3])
assert(ctd[1].list and ctd[2].list and ctd[3].list)
--@no-debug@

		local n = new()
		for tab = 1, numTabs do
			local branchLength = GetNumTalents(tab, isnotplayer, nil, group)
			if (branchLength ~= #ctd[tab].list) then
				-- Tab tree size is not what we expected for this class
				del(n)
				return
			end

			local t = new()
			local trim
			for i = 1,branchLength do
				local name, icon, tier, column, currentRank, maxRank = GetTalentInfo(tab, i, isnotplayer, nil, group)
				tinsert(t, currentRank)
				if (currentRank > 0) then
					trim = i			-- We strip off trailing zeros from talent strings to save storage space
				end
			end

			tinsert(n, table.concat(t, nil, 1, trim or 0))
			del(t)
		end

		return n
	end
end

-- TalentQuery_Ready
function lib:TalentQuery_Ready_Outsider(e, name, realm, unit)
	self:TalentQuery_Ready(e, name, realm, unit)
end

-- TalentQuery_Ready
function lib:TalentQuery_Ready(e, name, realm, unit)
--d("TalentQuery_Ready(%s, %s, %s)", name, realm or "nil", unit)
	local guid = unit and UnitGUID(unit)
	local r = guid and self.roster[guid]
	if (r) then
		local namerealm = realm and realm ~= "" and name .. "-" .. realm or name
		local isnotplayer = not UnitIsUnit(unit, "player")

		if (GetTalentTabInfo(1, isnotplayer)) then
			self:GetClassTalentData(unit)

			local active = GetActiveTalentGroup(isnotplayer)
			local numActive = GetNumTalentGroups(isnotplayer)
			local listUnspent
			local talents, alternateTalents
			local invalid

			for group = 1,numActive do
				local n = ReadTalentGroup(isnotplayer, group, r.class)
				if (n) then
					if (active == group) then
						talents = n
					else
						if (not alternateTalents) then
							alternateTalents = new()
						end
						alternateTalents[group] = n
					end
				else
					invalid = tree
					break
				end

				local unspent = GetUnspentTalentPoints(isnotplayer, nil, group)
				if (unspent and unspent > 0) then
					if (not listUnspent) then
						listUnspent = new()
					end
					listUnspent[group] = unspent
				end
			end

			if (isnotplayer and (invalid or (listUnspent and listUnspent[active] or 0) > 0)) then
				-- Unit didn't have all their points spent in active group, so we'll have another look in 10 seconds
				-- Don't need to check when it's "player" because we get PLAYER_TALENT_UPDATE event on changes
				self:TriggerRefreshTalents(guid, 10)
			end

			if (not invalid) then
				self:OnReceiveTalents(guid, unit, talents, alternateTalents, active, numActive, listUnspent)
			end
		end
	end
end
TalentQuery.RegisterCallback(lib, "TalentQuery_Ready")
TalentQuery.RegisterCallback(lib, "TalentQuery_Ready_Outsider")

-- GetUnitTalentSpec
function lib:GetUnitTalentSpec(unit, group)
	return self:GetGUIDTalentSpec(UnitGUID(unit), group)
end

-- GetGUIDTalentSpec
function lib:GetGUIDTalentSpec(guid, group)
	local talents, r = GetGUIDTalentsRaw(guid, group)
	if (talents) then
		return TalentWeight(talents, r.class)
	end
end

-- CompareTalents
local function CompareTalents(tree1, tree2)
	if ((tree1 ~= nil) ~= (tree2 ~= nil)) then
		return
	end
	if (tree1 and tree2 and #tree1 == #tree2) then
		for i = 1,#tree1 do
			if (tree1[i] ~= tree2[i]) then
				return
			end
		end
		return true
	end
end

-- OnReceiveTalents
function lib:OnReceiveTalents(guid, unit, talents, alternateTalents, active, numActive, listUnspent)
	local r = self.roster[guid]
	if (r) then
		--d("OnReceiveTalents(%s, %s, %s/%s/%s)", guid or "nil", unit or "nil", talents[1], talents[2], talents[3])
		if (active ~= r.active or numActive ~= r.numActive or not CompareTalents(talents, r.talents)) then
			if (r.talents) then
				del(r.oldTalents)
				r.oldTalents = r.talents
			end
			deepDel(r.alternateTalents)
			del(r.unspent)

			r.alternateTalents = alternateTalents
			r.talents = talents
			r.active = active
			r.numActive = numActive
			r.unspent = listUnspent

			local newSpec, n1, n2, n3 = TalentWeight(r.talents, r.class)

			local fired
			if (r.oldTalents) then
				-- For those cases when we didn't have the alternate talents for a player for whatever reason.
				-- Maybe they just picked up dual talent spec, or gated to trainer to respec.
				local oldSpec, o1, o2, o3 = TalentWeight(r.oldTalents, r.class)

				if (o1 ~= n1 or o2 ~= n2 or o3 ~= n3) then
					--d("Fire: LibGroupTalents_Update(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)", guid or "nil", unit or "nil", newSpec or "nil", n1 or "nil", n2 or "nil", n3 or "nil", oldSpec or "nil", o1 or "nil", o2 or "nil", o3 or "nil")
					self.events:Fire("LibGroupTalents_Update", guid, unit, newSpec, n1, n2, n3, oldSpec, o1, o2, o3)
					fired = true
				end
			end

			if (not fired) then
				--d("Fire: LibGroupTalents_Update(%s, %s, %s, %s, %s, %s)", guid or "nil", unit and UnitName(unit) or "nil", newSpec or "nil", n1 or "nil", n2 or "nil", n3 or "nil")
				self.events:Fire("LibGroupTalents_Update", guid, unit, newSpec, n1, n2, n3)
			end
			self:GetGUIDRole(guid, true)

			tinsert(self.batch, guid)
			CheckForUpdateComplete()

			r.oldTalents = del(r.oldTalents)
			return
		end
	end
	del(talents)
end

-- OnReceiveGlyphs
function lib:OnReceiveGlyphs(guid, sender, glyphs)
	local r = self.roster[guid]
	if (r) then
		local oldGlyphs
		if (r.glyphs) then
			oldGlyphs = r.glyphs[r.active]
			r.glyphs = del(r.glyphs)
		end

		local all = new(strsplit(";", glyphs))
		for i,page in ipairs(all) do
			local list = new(strsplit(",", page))
			local tab = tonumber(tremove(list, 1))
			if (tab and #list >= 1) then
				if (not r.glyphs) then
					r.glyphs = new()
				end
				r.glyphs[tab] = table.concat(list, ",")
			end
			del(list)
		end
		del(all)

		local newGlyphs = r.glyphs and r.glyphs[r.active]
		if (newGlyphs ~= oldGlyphs) then
			--d("Fire: LibGroupTalents_GlyphUpdate(%s, %s)", guid or "nil", unit and UnitName(unit) or "nil")
			self.events:Fire("LibGroupTalents_GlyphUpdate", guid, unit)
		end
	else
		del(glyphs)
	end
end

-- GetUnitGlyphs
function lib:GetUnitGlyphs(unit, group)
	return self:GetGUIDGlyphs(UnitGUID(unit), group)
end

-- GetGUIDGlyphs
function lib:GetGUIDGlyphs(guid, group)
	local r = self.roster[guid]
	if (r) then
		local g = r.glyphs[group or r.active]
		if (g) then
			local temp = new(strsplit(",", g))
			for i,str in ipairs(temp) do
				temp[i] = tonumber(str)
			end
			local a, b, c, d, e, f = unpack(temp)
			del(temp)
			return a, b, c, d, e, f
		end
	end
end
                                        
-- UnitHasGlyph
function lib:UnitHasGlyph(unit, glyphID, group)
	return lib:GUIDHasGlyph(UnitGUID(unit), glyphID, group)
end

-- GUIDHasGlyph
function lib:GUIDHasGlyph(guid, glyphID, group)
	local ret
	local r = self.roster[guid]
	if (r) then
		local g = r.glyphs[group or r.active]
		if (g) then
			local temp = new(strsplit(",", g))
			for i,str in ipairs(temp) do
				local id = tonumber(str)
				if (type(glyphID) == "number") then
					if (glyphID == id) then
						ret = true
						break
					end
				else
					if (glyphID == GetSpellInfo(id)) then
						ret = true
						break
					end
				end
			end
			del(temp)
		end
	end
	return ret
end

-- GLYPH_ADDED
function lib:GLYPH_ADDED(index, a, b, c)
	--d("GLYPH_ADDED(%s, %s, %s, %s)", tostring(index), tostring(a), tostring(b), tostring(c))
	self:RefreshPlayerGlyphs()
end

-- GLYPH_REMOVED
function lib:GLYPH_REMOVED(index, a, b, c)
	--d("GLYPH_REMOVED(%s, %s, %s, %s)", tostring(index), tostring(a), tostring(b), tostring(c))
	self:RefreshPlayerGlyphs()
end

-- GLYPH_UPDATED
function lib:GLYPH_UPDATED(index, a, b, c)
	--d("GLYPH_UPDATED(%s, %s, %s, %s)", tostring(index), tostring(a), tostring(b), tostring(c))
	self:RefreshPlayerGlyphs()
end

-- RefreshPlayerGlyphs
function lib:RefreshPlayerGlyphs()
	local guid = UnitGUID("player")
	local r = self.roster[guid]
	if (not r) then
		return
	end

	local glyphs = new()
	local any
	for talentGroup = 1,GetNumTalentGroups() do
		local list = new()
		for i = 1,GetNumGlyphSockets() do
			local enabled, glyphType, glyphSpell, icon = GetGlyphSocketInfo(i, talentGroup)
			if (enabled and glyphType and glyphSpell) then
				tinsert(list, glyphSpell)
				any = true
			end
		end
		glyphs[talentGroup] = table.concat(list, ",")
		del(list)
	end

	local oldGlyphs = r.glyphs
	if (any) then
		r.glyphs = glyphs
	end

	local change = (oldGlyphs and oldGlyphs[r.active]) ~= (r.glyphs and r.glyphs[r.active])
	if (change) then
		self:SendMyGlyphs()
		--d("Fire: LibGroupTalents_GlyphUpdate(%s, %s)", guid or "nil", unit and UnitName(unit) or "nil")
		self.events:Fire("LibGroupTalents_GlyphUpdate", guid, "player")
	end

	del(oldGlyphs)
end

-- PLAYER_TALENT_UPDATE
function lib:PLAYER_TALENT_UPDATE()
	self:TriggerRefreshTalents(UnitGUID("player"), 2)
end

-- UNIT_SPELLCAST_SUCCEEDED
function lib:UNIT_SPELLCAST_SUCCEEDED(unit, spell)
	local newActiveGroup = specChangers[spell]
	if (newActiveGroup) then
		local guid = UnitGUID(unit)
		local r = guid and self.roster[guid]
		if (r) then
			if (newActiveGroup == r.active) then
				-- We obviously didn't see them switch from this set
				self:GetGUIDRole(guid, true)
				return
			end

			local newSet = r.alternateTalents and r.alternateTalents[newActiveGroup]
			if (newSet) then
				-- We already have the alternate talents, so just swap them over with the active talents and notify
				r.alternateTalents[r.active] = r.talents
				r.talents = newSet
				r.active = newActiveGroup
				r.alternateTalents[newActiveGroup] = nil

				local oldSpec, o1, o2, o3 = TalentWeight(r.alternateTalents[newActiveGroup == 1 and 2 or 1], r.class)
				local newSpec, n1, n2, n3 = TalentWeight(r.talents, r.class)

				if (oldSpec ~= newSpec or o1 ~= n1 or o2 ~= n2 or o3 ~= n3) then
					--d("Fire: LibGroupTalents_Update(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)", guid or "nil", unit and UnitName(unit) or "nil", newSpec or "nil", n1 or "nil", n2 or "nil", n3 or "nil", oldSpec or "nil", o1 or "nil", o2 or "nil", o3 or "nil")
					self.events:Fire("LibGroupTalents_Update", guid, unit, newSpec, n1, n2, n3, oldSpec, o1, o2, o3)
				else
					--d("Fire: LibGroupTalents_Update(%s, %s, %s, %s, %s, %s)", guid or "nil", unit and UnitName(unit) or "nil", newSpec or "nil", n1 or "nil", n2 or "nil", n3 or "nil")
					self.events:Fire("LibGroupTalents_Update", guid, unit, newSpec, n1, n2, n3)
				end
				self:GetGUIDRole(guid, true)
				return
			end

			-- If we get this far, then something was wrong with the previously scanned talents, and we'll just rescan
			self:RefreshTalentsByGUID(guid)
		end
	end
end

-- TriggerRefreshTalents
function lib:TriggerRefreshTalents(guid, delay)
	if (not self.talentTimers) then
		self.talentTimers = new()
	end
	if (guid) then
		self.talentTimers[guid] = GetTime() + delay
		frame:Show()
	end
end

-- RefreshTalentsByUnit
function lib:RefreshTalentsByUnit(unit)
	local guid = UnitGUID(unit)
	if (guid) then
		self:RefreshTalentsByGUID(guid)
	end
end

-- RefreshTalentsByGUID
function lib:RefreshTalentsByGUID(guid)
	local r = self.roster[guid]
	if (not r) then
		return
	end
	if (not ValidateUnit(r, guid)) then
		return
	end

	if (self.talentTimers) then
		self.talentTimers[guid] = nil
	end

	if (not self.talentThrottle) then
		self.talentThrottle = {}
	end
	for guidThrottle,when in pairs(self.talentThrottle) do
		if (when < GetTime() - 5) then
			self.talentThrottle[guidThrottle] = nil
		elseif (guid == guidThrottle) then
			return
		end
	end
	self.talentThrottle[guid] = GetTime()

	if (self.commQueried) then
		self.commQueried[guid] = nil
		if (not next(self.commQueried)) then
			self.commQueried = del(self.commQueried)
		end
	end

	r.refresh = true
	self:CheckForMissingTalents()

	if (UnitGUID("player") == guid) then
		self:SendMyTalents()
	end
end

-- CheckForMissingTalents
function lib:CheckForMissingTalents()
	local any
	for guid,info in pairs(self.roster) do
		if (not info.talents or (not UnitIsVisible(info.name) and UnitExists(info.name)) or info.refresh) then
			any = true
			info.refresh = nil
			self:GetUnitTalents(info.unit, true)
		end
	end

	if (any) then
		lib.refreshCheckTimer = 15
		frame:Show()
	end
end

do
	local bearForm = GetSpellInfo(5487)						-- Bear Form
	local direBearForm = GetSpellInfo(9634)					-- Dire Bear Form
	local frostPresence = GetSpellInfo(48263)				-- Frost Presence

	-- GetUnitRole
	function lib:GetUnitRole(unit, reset)
		local guid = UnitGUID(unit)
		if (guid) then
			return self:GetGUIDRole(guid, reset)
		end
	end

	-- GetGUIDRole
	function lib:GetGUIDRole(guid, reset)
		local r = guid and self.roster[guid]
		if (not r) then
			return
		end
		local unit = r.unit
		if (r.role and not reset) then
			return r.role
		end
		if (not ValidateUnit(r, guid)) then
			return
		end

		local class = r.class
		local role = nil

		if (class == "ROGUE" or class == "HUNTER") then
			role = "melee"
		elseif (class == "MAGE" or class == "WARLOCK") then
			role = "caster"
		elseif (class == "DEATHKNIGHT") then
			role = UnitAura(unit, frostPresence) and "tank" or "melee"
		else
			if (r.talents) then
				local specName, t1, t2, t3 = TalentWeight(r.talents, class)
				
				if (class == "PRIEST") then
					role = ((t1 + t2) > t3) and "healer" or "caster"
				elseif (class == "WARRIOR") then
					role = ((t1 + t2) > t3) and "melee" or "tank" 
				else
					local heavy = (t1 > t2 and t1 > t3 and 1) or (t2 > t1 and t2 > t3 and 2) or (t3 > t1 and t3 > t2 and 3) or 0
					if (class == "PALADIN") then
						role = heavy == 1 and "healer" or heavy == 2 and "tank" or heavy == 3 and "melee"

					elseif (class == "DRUID") then
						if (heavy == 2) then
							if (UnitAura(unit, bearForm) or UnitAura(unit, direBearForm)) then
								role = "tank"
							else
								role = "melee"
							end
						else
							role = heavy == 1 and "caster" or "healer"
						end

					elseif (class == "SHAMAN") then
						role = heavy == 1 and "caster" or heavy == 2 and "melee" or heavy == 3 and "healer"
					end
				end
			end
		end

		local oldrole = r.role
		r.role = role

		if (oldrole and r.role ~= oldrole) then
			--d("Fire: LibGroupTalents_RoleChange(%s, %s, %s, %s)", guid or "nil", unit and UnitName(unit) or "nil", role or "nil", oldrole or "nil")
			self.events:Fire("LibGroupTalents_RoleChange", guid, unit, role, oldrole)
		end
		return role
	end
end

-- GetUnitTalents
function lib:GetUnitTalents(unit, refetch)
	local guid = UnitGUID(unit)
	if (not guid) then
		return
	end
	return self:GetGUIDTalents(guid, refetch)
end

-- GetGUIDTalents
function lib:GetGUIDTalents(guid, refetch)
	local r = self.roster[guid]
	if (not r) then
		return
	end
	if (not ValidateUnit(r, guid)) then
		return
	end

	local unit = r.unit
	local name, realm = UnitName(unit)

	if (r.talents) then
		-- If someone is out of sight, we won't catch their talent swap spell cast, so we'll invalidate them here and recheck talents
		if ((not UnitIsVisible(unit) and UnitIsConnected(unit)) or (self.outOfSight and self.outOfSight[guid])) then
			if (not r.version) then
				refetch = true
			end
		end
	end

	if (not r.talents or refetch) then
		if (UnitIsUnit("player", unit)) then
			self:RefreshPlayerGlyphs()
			self:TalentQuery_Ready(nil, name, realm, unit)

		elseif ((UnitInRaid(unit) or UnitInParty(unit)) and UnitIsConnected(unit)) then
			TalentQuery:Query(unit)

			if (not UnitIsVisible(unit) or not CanInspect(unit)) then
				if (r.version) then
					local namerealm = name
					if (realm and realm ~= "") then
						namerealm = name.."-"..realm
					end
					
					if (r.version) then
						if (not self.commQueried or not self.commQueried[guid]) then
							if (not self.commQueried) then
								self.commQueried = new()
							end
							self.commQueried[guid] = true
							self:SendCommMessage("REQUESTTALENTS", namerealm)
						end
					end
				end
			end
		end

		if (self.outOfSight) then
			self.outOfSight[guid] = nil
		end
	end

	return r.talents
end

-- SendCommMessage
function lib:SendCommMessage(msg, target, channel)
	if (msg) then
		if (ChatThrottleLib) then
			ChatThrottleLib:SendAddonMessage("NORMAL", MAJOR, msg, channel or "WHISPER", target)
		else
			SendAddonMessage(MAJOR, msg, channel or "WHISPER", target)
		end
	end
end

-- Throttle - Purposely local to here
-- Abuse prevention. Yes, who would abuse addon comms? Noone would make a macro to crash a mod user would they. Right?
-- Well, this one time, at band camp. Someone thought it was super funny to make a macro that DCd PallyPower users
local function Throttle(sender, key)
	if (not throttle) then
		throttle = {}
	end
	local s = throttle[sender]
	if (not s) then
		s = {}
		throttle[sender] = s
	end

	local ret = (s[key] or 0) < GetTime() - 5	-- Same message key only allowable once every 5 secs from 1 person
	s[key] = GetTime()
	return ret
end

-- CHAT_MSG_ADDON
function lib:CHAT_MSG_ADDON(prefix, msg, channel, sender)
	if (prefix == MAJOR) then
		if (sender == playerName) then
			return
		elseif (not UnitInRaid(sender) and not UnitInParty(sender)) then
			return
		end

		local guid = UnitGUID(sender)
		if (not guid) then
			return
		end
		local r = self.roster[guid]
		if (not r) then
			return
		end

		local cmd, str = msg:match("^(%a+) *(.*)$")
		if (not cmd) then
			return
		end

		if (cmd == "TALENTS") then
			-- Talents come in form of:
			-- TALENTS tree1,tree2,tree3;active,numActive[;altTree1,altTree2,altTree3[;...]]
			--d("Received: TALENTS "..str)
			local talents, alternateTalents, active, numActive
			local altCount = 0

			local bulk = new(strsplit(";", str))
			for i,str2 in ipairs(bulk) do
				local list = new(strsplit(",", str2))
				if (#list > 0) then
					if (not talents) then
						talents = list
					elseif (not active) then
						active, numActive = tonumber(list[1]), tonumber(list[2])
					else
						if (not alternateTalents) then
							alternateTalents = new()
						end
						local num = tonumber(tremove(list, 1))
						if (num) then
							altCount = altCount + 1
							alternateTalents[num] = list
						else
							invalid = true
							break
						end
					end
				else
					del(list)
					invalid = true
					break
				end
			end
			if (not invalid and talents and next(talents)) then
				if (not (active and numActive and alternateTalents and altCount == numActive - 1)) then
					active, numActive, alternateTalents = nil, nil, nil
				end
				self:OnReceiveTalents(guid, sender, talents, alternateTalents, active, numActive)
			else
				del(bulk)
				del(talents)
				deepDel(alternateTalents)
			end

		elseif (cmd == "GLYPHS") then
			local invalid
			local pages = new(strsplit(";", str))
			local glyphs = new()
			for page,info in ipairs(pages) do
				local list = strsplit(",", info)
				local tab = tonumber(tremove(list, 1))
				if (tab) then
					glyphs[tab] = table.concat(info, ",")
				else
					invalid = true
					break
				end
			end
			if (not invalid) then
				self:OnReceiveGlyphs(guid, sender, glyphs)
			end
			del(pages)

		elseif (cmd == "REQUESTTALENTS") then
			if (Throttle(sender, "REQUESTTALENTS")) then
				self:SendMyTalents(sender)
				self:SendMyGlyphs(sender)
			end

		elseif (cmd == "HELLO") then
			r.version = tonumber(str)
			if (channel ~= "WHISPER") then
				if (Throttle(sender, "HELLO")) then
					self:SendCommMessage("HELLO "..MINOR, sender)
				end
			end
		end
	end
end

-- SendMy
local function SendMy(sender, str)
	if (sender) then
		if (UnitIsConnected(sender)) then
			lib:SendCommMessage(str, sender)
		end
	else
		for guid,info in pairs(lib.roster) do
			if (info.version and UnitIsConnected(info.name)) then
				lib:SendCommMessage(str, info.name)
			end
		end
	end
end

-- SendMyTalents
function lib:SendMyTalents(sender)
	if (self:UserCount() > 0) then
		local r = self.roster[UnitGUID("player")]
		local t = r and r.talents
		if (t) then
			local str = format("TALENTS %s,%s,%s", t[1], t[2], t[3])

			local a = r.alternateTalents
			if (a and r.active and r.numActive) then
				str = format("%s;%s,%s", str, r.active, r.numActive)
				for i,data in pairs(a) do
					str = format("%s;%d,%s,%s,%s", str, i, data[1], data[2], data[3])
				end
			end
		end
		SendMy(sender, str)
	end
end

-- SendMyGlyphs
function lib:SendMyGlyphs(sender)
	if (self:UserCount() > 0) then
		local r = self.roster[UnitGUID("player")]
		if (r) then
			local str = "GLYPHS "
			local i = 1
			for tab,g in pairs(r.glyphs) do
				local temp = format("%d,%s", tab, g)
				str = str .. (i > 1 and ";" or "") .. temp
				i = i + 1
			end
			SendMy(sender, str)
		end
	end
end

-- UserCount
function lib:UserCount()
	local count = 0
	for guid,info in pairs(self.roster) do
		if (info.version and not UnitIsUnit("player", info.name)) then
			count = count + 1
		end
	end
	return count
end

-- UnitHasTalent
-- eg: lib:UnitHasTalent("player", GetSpellInfo(talentSpellID))
-- Returns: nil, or number of points spent into talent
-- If the talent group is not specified, then the active talent group is used
function lib:UnitHasTalent(unit, talentName, group)
	return unit and self:GUIDHasTalent(UnitGUID(unit), talentName, group)
end

-- GUIDHasTalent
-- Returns: nil, or number of points spent into talent
function lib:GUIDHasTalent(guid, talentName, group)
	local talents, r = GetGUIDTalentsRaw(guid, group)
	if (talents and r.class) then
		local data = self.classTalentData[r.class]
		if (data) then
			local info = data.list and data.list[talentName]
			if (info) then
				local str = talents[info.treeIndex]
				if (str) then
					local amount = (str:byte(info.index) or 48) - 48
					return (amount or 0) > 0 and amount or nil
				end
			end
		end
	end
end

-- GetClassTalentInfo
function lib:GetClassTalentInfo(class, talentName)
-- Returns: Max Rank, Icon, Tab, Tier, Column, Tree Index
	local data = self.classTalentData[class]
	if (data) then
		local info = data.list and data.list[talentName]
		if (info) then
			return info.maxRank, info.icon, info.treeIndex, info.column, info.tier, info.index
		end
	end
end

-- GetActiveTalentGroup
function lib:GetActiveTalentGroup(unit)
	if (UnitIsUnit(unit, "player")) then
		return GetActiveTalentGroup()
	else
		local guid = unit and UnitGUID(unit)
		local r = guid and self.roster[guid]
		return r and r.active or nil
	end
end

-- GetNumTalentGroups
function lib:GetNumTalentGroups(unit)
	if (UnitIsUnit(unit, "player")) then
		return GetNumTalentGroups()
	else
		local guid = unit and UnitGUID(unit)
		local r = guid and self.roster[guid]
		return r and r.numActive or nil
	end
end

-- GetTalentTabInfo
function lib:GetTalentTabInfo(unit, tab, group)
	if (UnitIsUnit(unit, "player")) then
		return GetTalentTabInfo(tab, nil, nil, group or GetActiveTalentGroup())
	else
		local guid = unit and UnitGUID(unit)
		local r = guid and self.roster[guid]
		if (r and r.class) then
			local ctd = self.classTalentData[r.class]
			if (ctd and tab >= 1 and tab <= #ctd) then
				local spec, c1, c2, c3 = self:GetGUIDTalentSpec(guid, group)
				return ctd[tab].name, ctd[tab].icon, tab == 1 and c1 or tab == 2 and c2 or c3, ctd[tab].background, 0
			end
		end
	end
end

-- GetNumTalents
function lib:GetNumTalents(unit, tab)
	if (UnitIsUnit(unit, "player")) then
		return GetNumTalents(tab)
	else
		local _, class = UnitClass(unit)
		if (class) then
			local ctd = self.classTalentData[class]
			if (ctd and tab >= 1 and tab <= #ctd) then
				return #ctd[tab].list
			end
		end
	end
end

-- GetTalentInfo
function lib:GetTalentInfo(unit, tab, index, group)
	if (UnitIsUnit(unit, "player")) then
		return GetTalentInfo(tab, index, nil, nil, group or GetActiveTalentGroup())
	else
		local _, class = UnitClass(unit)
		if (class) then
			local ctd = self.classTalentData[class]
			if (ctd and tab >= 1 and tab <= #ctd) then
				local info = ctd[tab].list[index]
				if (info) then
					local spent = self:UnitHasTalent(unit, info.name, group)
					return info.name, info.icon, info.tier, info.column, spent or 0, info.maxRank
				end
			end
		end
	end
end

-- GetNumTalentTabs
function lib:GetNumTalentTabs()
	return GetNumTalentTabs()
end

-- GetNumTalentTabs
function lib:GetUnspentTalentPoints(unit, group)
	if (UnitIsUnit(unit, "player")) then
		return GetUnspentTalentPoints(nil, nil, group)
	else
		local guid = unit and UnitGUID(unit)
		local r = guid and self.roster[guid]
		if (r) then
			return r.unspent and r.unspent[group or r.active or 1]
		end
	end
end

-- GetTalentCount
function lib:GetTalentCount()
	local count, missing = 0, 0
	for guid,info in pairs(self.roster) do
		if (info.talents) then
			count = count + 1
		else
			missing = missing + 1
		end
	end
	return count, missing
end

-- GetTalentMissingNames
function lib:GetTalentMissingNames()
	local list = new()
	for unit in self:IterateRoster() do
		local guid = UnitGUID(unit)
		local r = guid and self.roster[guid]
		if (not r or not r.talents) then
			tinsert(list, r.name)
		end
	end
	local ret
	if (next(list)) then
		ret = table.concat(list, ",")
	end
	del(list)
	return ret
end

-- PurgeAndRescanTalents
function lib:PurgeAndRescanTalents()
	if (self.roster) then
		for guid,info in pairs(self.roster) do
			info.talents = del(info.talents)
			info.alternateTalents = deepDel(info.alternateTalents)
			info.active = nil
			info.numActive = nil
		end
	end
	self:CheckForMissingTalents()
end

-- Roster iterator
do
	local function iter(t)
		local key = t.id
		local ret
		if (t.mode == "raid") then
			if (key > t.r) then
				del(t)
				return nil
			end
			ret = "raid"..key
		else
			if (key > t.p) then
				del(t)
				return nil
			end
			ret = key == 0 and "player" or "party"..key
		end
		t.id = key + 1
		return ret
	end

	-- IterateRoster
	function lib:IterateRoster()
		local t = new()
		if (GetNumRaidMembers() > 0) then
			t.mode = "raid"
			t.id = 1
			t.r = GetNumRaidMembers()
		else
			t.mode = "party"
			t.id = 0
			t.p = GetNumPartyMembers()
		end
		return iter, t
	end
end
